/*! ********************************************
 * EffectorExt.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_EffectorExt_h
#define INCLUDE_EffectorExt_h

#ifndef INCLUDE_GrayBox_h
#include "GrayBox.hpp"
#endif

#ifndef INCLUDE_Effector_h
#include "Effector.hpp"
#endif

namespace sora {
namespace graybox {
namespace efx {
/***********************************************
	Class	Stereo Delay Effector
************************************************/
class EFXStereoDelay : public EffectorBase, public IGrayBoxModule
{
private:
	// Ring Buffer
	int32 _ringBufferSizeBits = 15;
	int32 _ringBufferSize = 1 << _ringBufferSizeBits;
	int32 _ringBufferMask = _ringBufferSize - 1;
	float* _delayBufferL = nullptr;
	float* _delayBufferR = nullptr;
	//
	float _sampleRate = 44100.0f;
	// Delay
	int32 _wp = 0;  // WritePoint [frame]
	// Delay L
	float _feedBackL = 0.4f;  // 0 ~ 1.0
	int32 _rpL = 0;           // ReadPoint [frame]
	int32 _delayL = 0;        // [frame]
	float _delayTimeL = 0.0f; // [Sec]
	// Delay R
	float _feedBackR = 0.4f;
	int32 _rpR = 0;
	int32 _delayR = 0;
	float _delayTimeR = 0.0f;
// getter, setter
public:
	int32 getRingBufferSize() { return _ringBufferSize; }
	float getSampleRate() { return _sampleRate; }
	void setSampleRate(float rate);
	
	float getDelayFeedBackLevelL() { return _feedBackL; }
	float getDelayFeedBackLevelR() { return _feedBackR; }
	void setDelayFeedBackLevelL(float val) { _feedBackL = val; }
	void setDelayFeedBackLevelR(float val) { _feedBackR = val; }
	
	float getDelayTimeL() { return _delayTimeL; }
	float getDelayTimeR() { return _delayTimeR; }
	float setDelayTimeL(float val);
	float setDelayTimeR(float val);
	
	int32 getDelayFrameL() { return _delayL; }
	int32 getDelayFrameR() { return _delayR; }
	void setDelayFrameL(int32 val);
	void setDelayFrameR(int32 val);
	
	void setEffectParam(float mix = 0.2f,
		float delayTimeL = 0.25f, float feedBackLevelL = 0.4f,
		float delayTimeR = 0.35f, float feedBackLevelR = 0.4f);
// Base - EffectorBase ---
	void applyEffect(float* const& outL, float* const& outR, const int32& startFrame, const int32& sampleFrames) override;
	void resetParameters() override;
	void clearBuffer() override;
	void initialize() override;
// Interface - IfGrayBoxModule -----
	void updateSampleRate(float) override;
	void updateAudioChannels(int32) override {}
	void updateProcessBlockSize(int32) override {}
	
	void resetModule() override;
	void initiModule() override;
	void connectModule(GrayBox*) override;
	void disconnectModule() override;
	void updateBypassStatus(bool) override;
	
	void startProcess() override;
	void stopProcess() override;
	void suspendProcess() override;
	void resumeProcess() override;
	void process(ProcessData*) override;
// Constructor
	EFXStereoDelay();
	EFXStereoDelay(int32 bufferSizeBits = 15);
// Copy Constructor
	EFXStereoDelay(const EFXStereoDelay&);
	EFXStereoDelay& operator = (const EFXStereoDelay&);
// Move Constructor
	EFXStereoDelay(EFXStereoDelay&&)noexcept;
	EFXStereoDelay& operator = (EFXStereoDelay&&)noexcept;
// Destructor
	~EFXStereoDelay();
};
/***********************************************
	Class	Stereo Revervb Effector
************************************************/
class EFXStereoRevervb : public EffectorBase, public IGrayBoxModule
{
private:
	// Ring Buffer
	int32 _ringBufferSizeBits = 15;
	int32 _ringBufferSize = 1 << _ringBufferSizeBits;
	int32 _ringBufferMask = _ringBufferSize - 1;
	float* _delayBuffer1 = nullptr;
	float* _delayBuffer2 = nullptr;
	//
	float _sampleRate = 44100.0f;
	// Delay
	int32 _wp = 0;  // WritePoint [frame]
	// Delay 1
	float _feedBack1 = 0.5f;  // 0 ~ 1.0
	int32 _rp1 = 0;           // ReadPoint [frame]
	int32 _delay1 = 0;        // [frame]
	float _delayTime1 = 0.0f; // [Sec]
	// Delay 2
	float _feedBack2 = 0.3f;
	int32 _rp2 = 0;
	int32 _delay2 = 0;
	float _delayTime2 = 0.0f;
// getter, setter
public:
	int32 getRingBufferSize() { return _ringBufferSize; }
	float getSampleRate() { return _sampleRate; }
	void setSampleRate(float rate);
	
	float getDelayFeedBackLevel1() { return _feedBack1; }
	float getDelayFeedBackLevel2() { return _feedBack2; }
	void setDelayFeedBackLevel1(float val) { _feedBack1 = val; }
	void setDelayFeedBackLevel2(float val) { _feedBack2 = val; }
	
	float getDelayTime1() { return _delayTime1; }
	float getDelayTime2() { return _delayTime2; }
	float setDelayTime1(float val);
	float setDelayTime2(float val);

	int32 getDelayFrame1() { return _delay1; }
	int32 getDelayFrame2() { return _delay2; }
	void setDelayFrame1(int32 val);
	void setDelayFrame2(int32 val);
	
	void setEffectParam(float mix = 0.3f,
		float delayTime1 = 0.1f, float feedBackLevel1 = 0.5f,
		float delayTime2 = 0.25f, float feedBackLevel2 = 0.2f);
// Base - EffectorBase ---
	void applyEffect(float* const& outL, float* const& outR, const int32& startFrame, const int32& sampleFrames) override;
	void resetParameters() override;
	void clearBuffer() override;
	void initialize() override;
// Interface - IfGrayBoxModule -----
	void updateSampleRate(float) override;
	void updateAudioChannels(int32) override {}
	void updateProcessBlockSize(int32) override {}
	
	void resetModule() override;
	void initiModule() override;
	void connectModule(GrayBox*) override;
	void disconnectModule() override;
	void updateBypassStatus(bool) override;
	
	void startProcess() override;
	void stopProcess() override;
	void suspendProcess() override;
	void resumeProcess() override;
	void process(ProcessData*) override;
// Constructor
	EFXStereoRevervb();
	EFXStereoRevervb(int32 bufferSizeBits = 15);
// Copy Constructor
	EFXStereoRevervb(const EFXStereoRevervb&);
	EFXStereoRevervb& operator = (const EFXStereoRevervb&);
// Move Constructor
	EFXStereoRevervb(EFXStereoRevervb&&)noexcept;
	EFXStereoRevervb& operator = (EFXStereoRevervb&&)noexcept;
// Destructor
	~EFXStereoRevervb();
};
}}}
#endif