/*! ********************************************
 * GrayBox.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_GrayBox_h
#include "GrayBox.hpp"
#endif

#include <utility>

namespace sora {
namespace graybox {

/***********************************************
	Class	MidiMessage
************************************************/
void MidiMessage::clear()
{
	type = 0x80;
	channel = 0;
	value0 = 0;
	value1 = 0;
	dataSize = 0;
}
/***********************************************
	Class	MidiEvent
************************************************/
void MidiEvent::clear()
{
	midiPorts = 0xFFFF;
	next = nullptr;
	offsetFrames = 0;
	type = 0x80;
	channel = 0;
	value0 = 0;
	value1 = 0;
	dataSize = 0;
}
/***********************************************
	Class	ProcessData
************************************************/
void ProcessData::addToMidiEventList(MidiEvent& event, bool check)
{
	MidiEvent* _list;
	// check ---
	if (check)
	{
		_list = midiEvents->next;
		MidiEvent* _evptr = &event;
		while (_list)
		{
			if (_list == _evptr) return;
			_list = _list->next;
		}
	}
	// add to list ---
	_list = midiEvents;
	while (_list->next != nullptr)
	{
		if (event.offsetFrames >= _list->next->offsetFrames) _list = _list->next;
		else break;
	}
	event.next = _list->next;
	_list->next = &event;
}
/***********************************************
	Class	IGrayBoxModule
************************************************/
IGrayBoxModule::~IGrayBoxModule() {}
/***********************************************
	Class	GrayBox
************************************************/
// Main System ---
float GrayBox::setSampleRate(float sampleRate)
{
	if (sampleRate < 1.0f)return _processData.sampleRate;
	_processData.sampleRate = sampleRate;
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr) _modules[i]->updateSampleRate(sampleRate);
	return _processData.sampleRate;
}
int32 GrayBox::setAudioChannels(int32 ch)
{
	if (ch < 2)return _processData.audioChannels;
	_processData.audioChannels = ch;
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr) _modules[i]->updateAudioChannels(ch);
	return _processData.audioChannels;
}
int32 GrayBox::setProcessBlockSize(int32 size)
{
	if (size < 1)return _processData.blockSize;
	_processData.blockSize = size;
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr) _modules[i]->updateProcessBlockSize(size);
	return _processData.audioChannels;
}
void GrayBox::initialize()
{
	_defAudioLatency = 0.0f;
	_audioLatency = 0.0f;
	_clearProcessData();
	for (int i = 0; i < _moduleSlots; ++i)
	{
		detachModule(i);
		_bypassFlag[i] = false;
	}
}
// Module ---
int32 GrayBox::setModuleSlots(int32 num)
{
	if (num < 1) return _moduleSlots;
	for (int i = 0; i < _moduleSlots; ++i)detachModule(i);
	delete[] _modules;
	delete[] _bypassFlag;
	_modules = new IGrayBoxModule*[num];
	_bypassFlag = new bool[num];
	for (int i = 0; i < _moduleSlots; ++i)
	{
		_modules[i] = nullptr;
		_bypassFlag[i] = false;
	}
	return _moduleSlots = num;
}
IGrayBoxModule* GrayBox::getModule(int32 index)
{
	if (index >= _moduleSlots) return nullptr;
	return _modules[index];
}
IGrayBoxModule* GrayBox::attachModule(int32 index, IGrayBoxModule* module)
{
	if (index >= _moduleSlots) return nullptr;
	IGrayBoxModule* _mdl;
	for (int i = 0; i < _moduleSlots; ++i)
	{
		if (module != _modules[i]) continue;
		_mdl = _modules[i];
		_modules[i] = nullptr;
		_mdl->disconnectModule();
	}
	_modules[index] = module;
	module->connectModule(this);
	return module;
}
IGrayBoxModule* GrayBox::detachModule(int32 index)
{
	if (index >= _moduleSlots) return nullptr;
	IGrayBoxModule* _mdl = _modules[index];
	_modules[index] = nullptr;
	if (_mdl != nullptr)_mdl->disconnectModule();
	return _mdl;
}
bool GrayBox::getBypassStatus(int32 index)
{
	if (index >= _moduleSlots) return false;
	return _bypassFlag[index];
}
void GrayBox::bypassProcess(int32 index, bool bypass)
{
	if (index >= _moduleSlots) return;
	_bypassFlag[index] = bypass;
	if (_modules[index] != nullptr)_modules[index]->updateBypassStatus(bypass);
}
void GrayBox::resetModules()
{
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr)_modules[i]->resetModule();
}
void GrayBox::initiModules()
{
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr)_modules[i]->initiModule();
}
// Process ---
void GrayBox::suspendProcess()
{
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr)_modules[i]->suspendProcess();
}
void GrayBox::resumeProcess()
{
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr)_modules[i]->resumeProcess();
}
void GrayBox::startProcess()
{
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr)_modules[i]->startProcess();
}
void GrayBox::stopProcess()
{
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr)_modules[i]->stopProcess();
}
void GrayBox::_clearProcessData()
{
	_processData.midiEvents = nullptr;
	_processData.audioOut = nullptr;
	_processData.offsetFrames = 0;
	_processData.sampleFrames = 0;
	_audioLatency = _defAudioLatency;
}
void GrayBox::process(float** audioOut, int32 offsetFrames, int32 sampleFrames)
{
// begin Process -->
	_eventList.clear();
	_processData.midiEvents = &_eventList;
	_processData.audioOut = audioOut;
	_processData.offsetFrames = offsetFrames;
	_processData.sampleFrames = sampleFrames;
	_processData.audioLatency = _audioLatency;
// Modules Process -->
	for (int i = 0; i < _moduleSlots; ++i) if (_modules[i] != nullptr && !_bypassFlag[i]) _modules[i]->process(&_processData);
// after Process -->
	_clearProcessData();
}
// Constructor
GrayBox::GrayBox(int32 blockSize, float sampleRate, int32 audioChannels, int32 moduleSlots)
{
	// processData
	_processData.blockSize = (blockSize < 1)? 1 : blockSize;
	_processData.sampleRate = (sampleRate < 1.0f)? 1.0f : sampleRate;
	_processData.audioChannels = (audioChannels >= 2)? audioChannels : 2;
	// module
	_moduleSlots = (moduleSlots < 1)? 1 : moduleSlots;
	_modules = new IGrayBoxModule*[moduleSlots];
	_bypassFlag = new bool[moduleSlots];
	for (int i = 0; i < _moduleSlots; ++i)
	{
		_modules[i] = nullptr;
		_bypassFlag[i] = false;
	}
}
GrayBox::GrayBox(GrayBox && x) noexcept
{
	_eventList = std::move(_eventList);
	_processData = std::move(_processData);
	_moduleSlots = x._moduleSlots;
	_modules = x._modules;
	_bypassFlag = x._bypassFlag;
	_defAudioLatency = x._defAudioLatency;
	_audioLatency = x._audioLatency;
	x._modules = nullptr;
	x._bypassFlag = nullptr;
}
GrayBox & GrayBox::operator=(GrayBox && x) noexcept
{
	if (this == &x)return *this;
	_eventList = std::move(_eventList);
	_processData = std::move(_processData);
	_moduleSlots = x._moduleSlots;
	if (_modules != nullptr)delete[] _modules;
	_modules = x._modules; x._modules = nullptr;
	if (_bypassFlag != nullptr)delete[] _bypassFlag;
	_bypassFlag = x._bypassFlag; x._bypassFlag = nullptr;
	_defAudioLatency = x._defAudioLatency;
	_audioLatency = x._audioLatency;
	return *this;
}
GrayBox::~GrayBox()
{
	if (_modules != nullptr)delete[] _modules;
	_modules = nullptr;
	if (_bypassFlag != nullptr)delete[] _bypassFlag;
	_bypassFlag = nullptr;
}
}}