/*! ********************************************
 * MidiController.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_MidiController_h
#include "MidiController.hpp"
#endif

namespace sora {
namespace graybox {
namespace util {

void MidiController::clearMIDIEventBuffer()
{
	for (int i = 0; i < _eventBufferSize; ++i)_eventBuffer[i].clear();
	_usedEvents = 0;
}
int32 MidiController::setEventBufferSize(int32 size)
{
	if (size == _eventBufferSize) return _eventBufferSize;
	if (size < 64)size = 64;
	delete[] _eventBuffer;
	_eventBuffer = new MidiEvent[_eventBufferSize];
	_usedEvents = 0;
	return _eventBufferSize = size;
}
void MidiController::noteOff(const int32 & ch, const int32 & key, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0x80;
	_ev.channel = ch;
	_ev.value0 = key;
}
void MidiController::noteOn(const int32 & ch, const int32 & key, const int32 & velo, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0x90;
	_ev.channel = ch;
	_ev.value0 = key;
	_ev.value1 = velo;
}
void MidiController::keyAfterTouch(const int32 & ch, const int32 & key, const int32 & val, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xA0;
	_ev.channel = ch;
	_ev.value0 = key;
	_ev.value1 = val;
}
void MidiController::controlChange(const int32 & ch, const int32 & ccn, const int32 & val, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xB0;
	_ev.channel = ch;
	_ev.value0 = ccn;
	_ev.value1 = val;
}
void MidiController::programChange(const int32 & ch, const int32 & val, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xC0;
	_ev.channel = ch;
	_ev.value0 = val;
}
void MidiController::channelAfterTouch(const int32 & ch, const int32 & val, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xD0;
	_ev.channel = ch;
	_ev.value0 = val;
}
void MidiController::pitchBendChange(const int32 & ch, const int32 & lsb, const int32 & msb, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xE0;
	_ev.channel = ch;
	_ev.value0 = lsb;
	_ev.value1 = msb;
}
void MidiController::pitchBendChangeW(const int32 & ch, const int32 & val, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xE0;
	_ev.channel = ch;
	_ev.value0 = val & 0x7f;
	_ev.value1 = val >> 7 & 0x7f;
}
void MidiController::systemExclusive(const uint8* msg, const int32& dataSize, const int32& offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = 0xF0;
	_ev.channel = 0;
	int _size = (dataSize > MidiEvent::DATA_PACKET_SIZE) ? MidiEvent::DATA_PACKET_SIZE : dataSize;
	_ev.dataSize = _size;
	for (int i = 0; i < _size; ++i)_ev.data[i] = msg[i];
}
void MidiController::midiMessage(
		const int32 & type, const int32 & ch, const int32 & val0, const int32 & val1, 
		const uint8* msg, const int32& dataSize, const int32 & offsetFrames)
{
	if (_usedEvents >= _eventBufferSize) return;
	MidiEvent& _ev = _eventBuffer[_usedEvents++];
	_ev.next = nullptr;  // clear
	_ev.midiPorts = midiOutPorts;
	_ev.offsetFrames = offsetFrames;
	_ev.type = type;
	_ev.channel = ch;
	_ev.value0 = val0;
	_ev.value1 = val1;
	if (msg == nullptr) { _ev.dataSize = 0u; return; }
	int _size = (dataSize > MidiEvent::DATA_PACKET_SIZE) ? MidiEvent::DATA_PACKET_SIZE : dataSize;
	_ev.dataSize = _size;
	for (int i = 0; i < _size; ++i)_ev.data[i] = msg[i];
}
// Constructor
MidiController::MidiController()
{
	_eventBuffer = new MidiEvent[_eventBufferSize];
	_usedEvents = 0;
}
MidiController::MidiController(int32 eventBufferSize)
{
	if (eventBufferSize < 16)eventBufferSize = 16;
	_eventBufferSize = eventBufferSize;
	_eventBuffer = new MidiEvent[_eventBufferSize];
	_usedEvents = 0;
}
// Move Constructor
MidiController::MidiController(MidiController&& x)noexcept
{
	_eventBuffer = x._eventBuffer; x._eventBuffer = nullptr;
	_eventBufferSize = x._eventBufferSize;
	_usedEvents = x._usedEvents;
	midiOutPorts = x.midiOutPorts;
}
MidiController& MidiController::operator = (MidiController&& x)noexcept
{
	if (this == &x)return *this;
	if (_eventBuffer != nullptr)delete[] _eventBuffer;
	_eventBuffer = x._eventBuffer; x._eventBuffer = nullptr;
	_eventBufferSize = x._eventBufferSize;
	_usedEvents = x._usedEvents;
	midiOutPorts = x.midiOutPorts;
	return *this;
}
// Destructor
MidiController::~MidiController()
{
	if (_eventBuffer != nullptr)delete[] _eventBuffer;
	_eventBuffer = nullptr;
}
// Interface - IfGrayBoxModule -----
void MidiController::resetModule(){ clearMIDIEventBuffer(); }
void MidiController::initiModule(){ clearMIDIEventBuffer(); }
void MidiController::connectModule(GrayBox*){ clearMIDIEventBuffer(); }
void MidiController::disconnectModule(){ clearMIDIEventBuffer(); }
void MidiController::updateBypassStatus(bool) { clearMIDIEventBuffer(); }

void MidiController::startProcess(){ clearMIDIEventBuffer(); }
void MidiController::stopProcess(){ clearMIDIEventBuffer(); }
void MidiController::suspendProcess(){ clearMIDIEventBuffer(); }
void MidiController::resumeProcess(){ clearMIDIEventBuffer(); }
void MidiController::process(ProcessData* procData)
{
	MidiEvent* _list, *_ev;
	while (_usedEvents > 0)
	{
		_ev = &_eventBuffer[--_usedEvents];
		_list = procData->midiEvents;
		while (_list->next)
		{
			if (_ev->offsetFrames >= _list->next->offsetFrames) _list = _list->next;
			else break;
		}
		_ev->next = _list->next;
		_list->next = _ev;
	}
}
}}}