/*! ********************************************
 * SynthesizerPCM.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerPCM_h
#include "SynthesizerPCM.hpp"
#endif

#include <utility>
#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	SynthesizerPCM
************************************************/
void SynthesizerPCM::_recreateVU(int32 maxPoly)
{
	// Recreate Array
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = new VoiceUnitBase*[maxPoly];
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = new VoiceUnitPCM[maxPoly];
	// Reset VoiceUnit
	for (int i = 0; i < maxPoly; ++i)
	{
		VoiceUnitPCM& _vu = _voiceUnitExt[i];
		_voiceUnits[i] = &_voiceUnitExt[i];
		_vu.killProcess();
		_vu.VOICE_PARTS = VOICE_PARTS;
		_vu.SAMPLE_DATAS = SAMPLE_DATAS;
		_vu.TONE_PRESETS = TONE_PRESETS;
		_vu.SAMPLER_PRESETS = SAMPLER_PRESETS;
		_vu.setMasterVolume(_masterVolume);
	}
	// set Property
	_maxPolyphony = maxPoly;
}
void SynthesizerPCM::_resetKeyParams()
{
	const double _sampleRateD = double(_sampleRate);
	// LFO
	double _lfoWaveSize = SynthData::LFO_WAVE_TABLE_SIZE / 2.0;
		// Max Frame Rate - LFO Phase Step
	double _lfoMaxFrameRate = _sampleRateD / SynthData::LFO_MAX_FREQ / _lfoWaveSize;
	if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
	int _lfoTimerCoef = int(SynthData::LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		
	// EG
	double _rate0Frames = SynthData::EG_REFERENCE_TIME * SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egPow2tIdx = SynthData::toPow2tIndexValue(SynthData::EG_MAX_CNT / _rate0Frames);
	double _sndOffFrames = SynthData::EG_SOUND_OFF_TIME / SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egSndOff = int(SynthData::EG_MAX_CNT / _sndOffFrames + 0.5);
	// PG
	int _smpl = SynthData::toPow2tIndexValue(_sampleRateD);
	int _pgSamplePow2tIdx = SynthData::toPow2tIndexValue(double(1 << PG_PHASE_FP)) - _smpl;
	// PG WaveTable
	int _pgKeyPow2tIdx = _pgSamplePow2tIdx + SynthData::toPow2tIndexValue(double(_referencePitch) * std::pow(2.0, -69.0 / 12.0)); // note 0 Pitch
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitPCM& _vu = _voiceUnitExt[i];
		_vu.killProcess();
		_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
		_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
		_vu.egPow2tIdx = _egPow2tIdx;
		_vu.egPhaseStepSndOff = _egSndOff;
		_vu.lfoTimerCoef = _lfoTimerCoef;
	}
}
void SynthesizerPCM::setSampleData(int32 id, const SampleData& sampleData, bool deepCopy)
{
	allSoundStop();
	SAMPLE_DATAS[id].copy(sampleData, deepCopy);
}
void SynthesizerPCM::clearSampleData(int32 id)
{
	allSoundStop();
	SAMPLE_DATAS[id].clear();
}
void SynthesizerPCM::factoryDataReset()
{
	if (_voiceUnitExt != nullptr) allSoundStop();
	// Preset
	if (TONE_PRESETS == nullptr)TONE_PRESETS = new TonePresetPCM[128];
	for (int i = 0; i < 128; ++i)TONE_PRESETS[i].reset();
// Sampler
	if (SAMPLE_DATAS == nullptr)SAMPLE_DATAS = new SampleData[128];
	if (SAMPLER_PRESETS == nullptr)SAMPLER_PRESETS = new SamplerPreset[128];
	for (int i = 0; i < 128; ++i)
	{
		SAMPLE_DATAS[i].clear();
		SAMPLER_PRESETS[i].reset();
	}
}
// Constructor
SynthesizerPCM::SynthesizerPCM(int32 maxPoly, float refPitch, float masVol) : MultiTimberSynthBase::MultiTimberSynthBase(maxPoly, refPitch, masVol)
{
	factoryDataReset();
	_recreateVU(_maxPolyphony);
	_resetKeyParams();
}
SynthesizerPCM::SynthesizerPCM(SynthesizerPCM && x) noexcept : MultiTimberSynthBase::MultiTimberSynthBase(std::move(x))
{
	// Synthesizer - Data -----
	SAMPLE_DATAS = x.SAMPLE_DATAS; x.SAMPLE_DATAS = nullptr;
	TONE_PRESETS = x.TONE_PRESETS; x.TONE_PRESETS = nullptr;
	SAMPLER_PRESETS = x.SAMPLER_PRESETS; x.SAMPLER_PRESETS = nullptr;
	// Synthesizer - main -----
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;
}
SynthesizerPCM & SynthesizerPCM::operator=(SynthesizerPCM && x) noexcept
{
	if (this == &x)return *this;
	// MultiTimberSynthBase::operator=(std::move(x));
// MultiTimberSynthBase ===
	// Synth Status
	midiInPorts = x.midiInPorts;
	midiOutPorts = x.midiOutPorts;
	receiveSysEx = x.receiveSysEx;
	// Channel Paramters
	if (VOICE_PARTS != nullptr)delete[] VOICE_PARTS;
	VOICE_PARTS = x.VOICE_PARTS; x.VOICE_PARTS = nullptr;
	_sampleRate = x._sampleRate;
	_referencePitch = x._referencePitch;
	_masterVolume = x._masterVolume;
	// VoiceUnits
	_maxPolyphony = x._maxPolyphony;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = x._voiceUnits; x._voiceUnits = nullptr;
	_nton = x._nton;
	_ntonM = x._ntonM;
	// RPN,NRPN -----
	_rpnCtrlFlag = x._rpnCtrlFlag;
	_nrpnCtrlFlag = x._nrpnCtrlFlag;
// SynthesizerPCM ===
	// Synthesizer - Data -----
	if (SAMPLE_DATAS != nullptr)delete[] SAMPLE_DATAS;
	SAMPLE_DATAS = x.SAMPLE_DATAS; x.SAMPLE_DATAS = nullptr;
	if (TONE_PRESETS != nullptr)delete[] TONE_PRESETS;
	TONE_PRESETS = x.TONE_PRESETS; x.TONE_PRESETS = nullptr;
	if (SAMPLER_PRESETS != nullptr)delete[] SAMPLER_PRESETS;
	SAMPLER_PRESETS = x.SAMPLER_PRESETS; x.SAMPLER_PRESETS = nullptr;
	// Synthesizer - main -----
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;

	return *this;
}
SynthesizerPCM::~SynthesizerPCM()
{
	if (VOICE_PARTS != nullptr)delete[] VOICE_PARTS;
	VOICE_PARTS = nullptr;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = nullptr;
	if (SAMPLE_DATAS != nullptr)delete[] SAMPLE_DATAS;
	SAMPLE_DATAS = nullptr;
	if (TONE_PRESETS != nullptr)delete[] TONE_PRESETS;
	TONE_PRESETS = nullptr;
	if (SAMPLER_PRESETS != nullptr)delete[] SAMPLER_PRESETS;
	SAMPLER_PRESETS = nullptr;
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = nullptr;
}
// MidiEvent ---
void SynthesizerPCM::_noteOn(const MidiEvent& ev)
{
	VoicePart& _pt = VOICE_PARTS[ev.channel];
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitPCM& _vu = _voiceUnitExt[i];
		if (_vu.state == 0)
		{
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	const int _len = _nton + _maxPolyphony;
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitPCM& _vu = _voiceUnitExt[i & _ntonM];
		// (KeyOff || SoundOff) -->
		if (_vu.state & 0x60)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitPCM& _vu = _voiceUnitExt[i & _ntonM];
		// KeyOn -->
		if (_vu.state & 0x10)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
}
void SynthesizerPCM::_readSysExData(const MidiEvent& ev)
{
	if (ev.dataSize < 8)return;
	const uint8* dat = ev.data;
	const uint8* dataEnd = dat + ev.dataSize - 1;
	if (*(dat++) != 0x7Du)return;  // ManuFactureID
	if (*(dat++) != 0x10u)return;  // DeviceID
	if (*(dat++) != 0x23u)return;  // ModelID

	switch (*(dat++))	// Address 1
	{
	// Common
		case 0x00u:
			if (*(dat++) != 0x00u)return; // Address 2
			switch (*(dat++))	// Address 3
			{
				case 0x00u : setMasterVolume(_MASTER_VOLUME_TABLE[*(dat++)]); if (dat >= dataEnd) break;
				case 0x01u : setReferencePitch(440.0f + (*dat - 64.0f) * 0.5f); break;
			}
			break;
	// Inst (Program map)
		case 0x04u: {
			TonePresetPCM& _preset = TONE_PRESETS[*(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _preset.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _preset.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _preset.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _preset.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _preset.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _preset.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _preset.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _preset.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _preset.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _preset.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _preset.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _preset.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _preset.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _preset.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _preset.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _preset.toneName[15] = *(dat++);  if (dat >= dataEnd) break;
				
				case 0x10u : dat++;  if (dat >= dataEnd) break;
				case 0x11u : dat++;  if (dat >= dataEnd) break;
				case 0x12u : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x13u : _preset.lfoFreq = _preset.lfoFreq & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x14u : _preset.strc    = *(dat++) & 3;                             if (dat >= dataEnd) break;
				case 0x15u : _preset.tune    = _preset.tune & 0x3f80 | *(dat++);         if (dat >= dataEnd) break;
				case 0x16u : _preset.tune    = _preset.tune & 0x007f | *(dat++) << 7;    if (dat >= dataEnd) break;
				case 0x17u : _preset.lms     = *(dat++) & 15;  if (dat >= dataEnd) break;
				case 0x18u : _preset.smplNum = *(dat++);       if (dat >= dataEnd) break;
				case 0x19u : _preset.tl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Au : _preset.ar      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Bu : _preset.dr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Cu : _preset.sl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Du : _preset.rr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Eu : _preset.kf      = *(dat++) & 3;   if (dat >= dataEnd) break;
				case 0x1Fu : _preset.kfStart = *dat;           break;
			}
			} break;
	// Percussion
		case 0x06u: {
			SamplerPreset& _smplPrst = SAMPLER_PRESETS[*(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _smplPrst.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _smplPrst.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _smplPrst.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _smplPrst.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _smplPrst.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _smplPrst.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _smplPrst.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _smplPrst.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _smplPrst.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _smplPrst.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _smplPrst.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _smplPrst.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _smplPrst.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _smplPrst.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _smplPrst.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _smplPrst.toneName[15] = *(dat++);  if (dat >= dataEnd) break;

				case 0x10u : _smplPrst.smplNum = *(dat++);       if (dat >= dataEnd) break;
				case 0x11u : _smplPrst.pan     = *(dat++);       if (dat >= dataEnd) break;
				case 0x12u : _smplPrst.octv    = *(dat++) & 15;  if (dat >= dataEnd) break;
				case 0x13u : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x14u : _smplPrst.tune    = _smplPrst.tune & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x15u : _smplPrst.tl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x16u : _smplPrst.ar      = *(dat++);       if (dat >= dataEnd) break;
				case 0x17u : _smplPrst.dr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x18u : _smplPrst.sl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x19u : _smplPrst.rr      = *dat;           break;
			}
			} break;
	}
}
/***********************************************
	Class	SynthesizerPCM :: VoiceUnitPCM
************************************************/
void SynthesizerPCM::VoiceUnitPCM::keyOff()
{
	// KeyOn --->
	if (state & 0x0010)
	{
		// ATTACK ->
		if (_egState == 1)
		{
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			if (_egCount >= _EG_MAX_CNT)_egCount = 0;
			else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			if (_egCount > _egCntEnvBtm){ killProcess(); return; }
		}
		// DECAY,SUSTAIN ->
		_egGate = _egCntEnvBtm;
		_egCountStep = _egCountStepRR;
		_egState = 8;  // Release
		state = 0x0021;
		// sampler ---
		_loopE = _plbkE;
		_loopF = _loopFk;
	}
}
void SynthesizerPCM::VoiceUnitPCM::soundOff()
{
	// KeyOn || KeyOff --->
	if (state & 0x0030)
	{
		// ATTACK ->
		if (_egState == 1)
		{
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			if (_egCount >= _EG_MAX_CNT)_egCount = 0;
			else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << (_EG_CNT_FP - 1);
			if (_egCount > _egCntEnvBtm){ killProcess(); return; }
		}
		// DECAY,SUSTAIN ->
		_egGate = _egCntEnvBtm;
		_egCountStep = egPhaseStepSndOff;
		_egState = 0;  // SoundOff, EG OFF
		partNumber = -1; // SoundOff
		state = 0x0041;
		// sampler ---
		_loopE = _plbkE;
		_loopF = _loopFk;
	}
}
int32 SynthesizerPCM::VoiceUnitPCM::_shiftEGState()
{
	switch (_egState)
	{
		// ATTACK ->
		case 1:
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
			// -> DECAY
			if (_egSL > 0)
			{
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				return _egState = 2;
			}
			// -> SUSTAIN
			_egGate = 0x7fffffff;
			_egCountStep = 0;
			return _egState = 4;
		// DECAY ->
		case 2:
			// -> SUSTAIN
			if ((_egTL + _egSL) < _EG_SUSTAIN_BTM)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				_egGate = 0x7fffffff;
				_egCountStep = 0;
				_egCount = _egCntEnvSus;
				return _egState = 4;
			}
		// DECAY,RELEASE,SoundOFF -> EG OFF
		default: return _egState = 0;
	}
}
void SynthesizerPCM::VoiceUnitPCM::keyOn(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0x0011;
	partNumber = ch;
	noteNumber = note;
// Part, Preset ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	TonePresetPCM& _preset = TONE_PRESETS[_part.programChange];
	SampleData& _sample = SAMPLE_DATAS[_preset.smplNum];
	if (_sample.sampleData == nullptr){ killProcess(); return; }
	_waveTable = _sample.sampleData;
	_plbkE = _sample.plbkE;
	_loopE = _sample.loopE;
	_loopF = _sample.loopF;
	_loopFk = _sample.loopFk;
// LFO ---
	_lfoModSens = _preset.lms;
	_lfoMod = _part.ccValues[1] * _lfoModSens;
	_lfoTimerStep = _preset.lfoFreq * lfoTimerCoef;
	_lfoTimer = _LFO_TIMER_INIT_VALUE;
	_lfoPhase = _preset.strc + 1 & 1;
	_lfoOutPM = 0;
	_lfoOutAM = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_prPan = 64;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// PG ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	_pgKey = _POW2T_KEY_TABLE[noteNumber] + (_preset.tune - 8192) + _sample.pgKey + pgKeyPow2tIdx;
	int v = _pgKey + _pgBend;// & 0x7fffffff;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase = _sample.plbkS;
	_pgPhaseAFP = 0;
// EG ---
	_egTL = _preset.tl;
	_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
	_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
	_egSL = _preset.sl;
	_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
	if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
	int _keyFlw = _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart] << _preset.kf;
	_keyFlw += egPow2tIdx;
	v = (_preset.dr << _EG_PRAM_OCTV_LS) + _keyFlw;
	_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset.rr << _EG_PRAM_OCTV_LS) + _keyFlw;
	_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset.ar < 127) {
		_egCount = _EG_CNT_ENV_ATK;
		_egLevelTable = _EG_LEVEL_TABLE_UP;
		_egGate = _EG_MAX_CNT;
		v = (_preset.ar << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState = 1;
	} else {
		_egLevelTable = _EG_LEVEL_TABLE_DW;
		_egCount = 0;
	// -> DECAY
		if (_egSL > 0) {
			_egGate = _egCntEnvSus;
			_egCountStep = _egCountStepDR;
			_egState = 2;
	// -> SUSTAIN
		} else {
			_egGate = 0x7fffffff;  // dummy
			_egCountStep = 0;  // dummy
			_egState = 4; }
	}
// process ---
	v = (_preset.strc & 3) << 1;
	if (_sample.channelCount == 2)v|=1;
	process = _PROC_PTRS[v];
}
void SynthesizerPCM::VoiceUnitPCM::keyOnSampler(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0x0011;
	partNumber = ch;
	noteNumber = note;
// tone ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	SamplerPreset& _preset = SAMPLER_PRESETS[noteNumber];
	SampleData& _sample = SAMPLE_DATAS[_preset.smplNum];
	if (_sample.sampleData == nullptr){ killProcess(); return; }
	_waveTable = _sample.sampleData;
	_plbkE = _sample.plbkE;
	_loopE = _sample.loopE;
	_loopF = _sample.loopF;
	_loopFk = _sample.loopFk;
// LFO ---
	_lfoOutPM = 0;
	_lfoOutAM = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_prPan = _preset.pan;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// PG ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	_pgKey = ((_preset.octv - 8) << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
	int v = _pgKey + _pgBend;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase = _sample.plbkS;
	_pgPhaseAFP = 0;
// EG ---
	_egTL = _preset.tl;
	_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
	_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
	_egSL = _preset.sl;
	_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
	if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
	v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset.ar < 127) {
		_egCount = _EG_CNT_ENV_ATK;
		_egLevelTable = _EG_LEVEL_TABLE_UP;
		_egGate = _EG_MAX_CNT;
		v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState = 1;
	} else {
		_egLevelTable = _EG_LEVEL_TABLE_DW;
		_egCount = 0;
	// -> DECAY
		if (_egSL > 0) {
			_egGate = _egCntEnvSus;
			_egCountStep = _egCountStepDR;
			_egState = 2;
	// -> SUSTAIN
		} else {
			_egGate = 0x7fffffff;  // dummy
			_egCountStep = 0;  // dummy
			_egState = 4; }
	}
// process ---
	process = (_sample.channelCount == 2)? _PROC_PTRS[7] : _PROC_PTRS[6];
}
void SynthesizerPCM::VoiceUnitPCM::updPartBendSens(const int32& val)
{
	_ptBendSens = _POW2T_KEY_TABLE[val];
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v = _pgKey + _pgBend + _lfoOutPM;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
}
void SynthesizerPCM::VoiceUnitPCM::updPartPitchBend(const int32& val)
{
	_ptBend = val - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v = _pgKey + _pgBend + _lfoOutPM;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
}
void SynthesizerPCM::VoiceUnitPCM::updPartModulation(const int32& val)
{
	_lfoMod = val * _lfoModSens;
}
void SynthesizerPCM::VoiceUnitPCM::updateLfoFrequency(const int32& val)
{
	_lfoTimerStep = val * lfoTimerCoef;
}
void SynthesizerPCM::VoiceUnitPCM::_proc0(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc1(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc2(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
			int l = _pgKey + _pgBend + _lfoOutPM;
			_pgPhaseStep = _POW2_TABLE[l & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[l >> _POW2T_OCTV_BITS];
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc3(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
			int l = _pgKey + _pgBend + _lfoOutPM;
			_pgPhaseStep = _POW2_TABLE[l & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[l >> _POW2T_OCTV_BITS];
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc4(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			int l = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;
			_lfoOutAM = l & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase] + v + _lfoOutAM;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc5(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			int l = (_LFO_WAVE_TABLE[_lfoPhase | 1] * _lfoMod) >> 11;
			_lfoOutAM = l & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv + _lfoOutAM;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc6(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerPCM::VoiceUnitPCM::_proc7(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
// Constructor
SynthesizerPCM::VoiceUnitPCM::VoiceUnitPCM() : VoiceUnitBase()
{
	SynthData* _synthData = SynthData::getInstance();
	// Log2 Volume Table
	_LOG2V_TABLE = _synthData->getLOG2V_TABLE();
	// Pow2 Table
	_POW2_TABLE      = _synthData->getPOW2_TABLE();
	_POW2T_RS_TABLE  = _synthData->getPOW2T_RS_TABLE();
	_POW2T_KEY_TABLE = _synthData->getPOW2T_KEY_TABLE();
	// EG
	_EG_TOTAL_LEVEL_TABLE = _synthData->getEG_TOTAL_LEVEL_TABLE();
	_EG_LEVEL_TABLE_UP    = _synthData->getEG_LEVEL_TABLE_UP();
	_EG_LEVEL_TABLE_DW    = _synthData->getEG_LEVEL_TABLE_DW();
	_EG_KEY_TABLE         = _synthData->getEG_KEY_TABLE();
	_EG_POW2T_RS_TABLE    = _synthData->getEG_POW2T_RS_TABLE();
	_EG_CNT_GATE_TABLE    = _synthData->getEG_CNT_GATE_TABLE();
	// etc
	_LFO_WAVE_TABLE = &_synthData->getLFO_WAVE_TABLES()[0];
}

}}}
