/*! ********************************************
 * SynthesizerWT.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerWT_h
#include "SynthesizerWT.hpp"
#endif

#include <utility>
#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	SynthesizerWT
************************************************/
void SynthesizerWT::_resetIndexTable()
{
	const double _RES = double(-SynthData::LOG2VT_DYN_RES),
		_BOADER = 1.0 / (1 << SynthData::LOG2VT_DYN_BITS);
	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	int32* _table = _INDEX_TABLE;
	for (int i = 0; i < 256; ++i)
	{
		double v = (i - 128) * 0.0078125;
		if      (v >  _BOADER) _table[i] =  int(std::log2( v) * _RES + 0.5) << 1;
		else if (v < -_BOADER) _table[i] = (int(std::log2(-v) * _RES + 0.5) << 1) + 1;
		else _table[i] = _BOTTOM;
	}
}
void SynthesizerWT::_recreateVU(int32 maxPoly)
{
	// Recreate Array
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = new VoiceUnitBase*[maxPoly];
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = new VoiceUnitWT[maxPoly];
	// Reset VoiceUnit
	for (int i = 0; i < maxPoly; ++i)
	{
		VoiceUnitWT& _vu = _voiceUnitExt[i];
		_voiceUnits[i] = &_voiceUnitExt[i];
		_vu.killProcess();
		_vu.VOICE_PARTS = VOICE_PARTS;
		_vu.WAVE_TABLES = WAVE_TABLES;
		_vu.TONE_PRESETS = TONE_PRESETS;
		_vu.SAMPLE_DATAS = SAMPLE_DATAS;
		_vu.SAMPLER_PRESETS = SAMPLER_PRESETS;
		_vu.pgPhaseMask = (_waveTableSize << PG_PHASE_FP) - 1;
		_vu.pgPhaseMaskNoise = _vu.pgPhaseMask >> 5;  // 32/1 update cycle
		_vu.setMasterVolume(_masterVolume);
	}
	// set Property
	_maxPolyphony = maxPoly;
}
void SynthesizerWT::_resetKeyParams()
{
	const double _sampleRateD = double(_sampleRate);
	// LFO
	double _lfoWaveSize = SynthData::LFO_WAVE_TABLE_SIZE / 2.0;
		// Max Frame Rate - LFO Phase Step
	double _lfoMaxFrameRate = _sampleRateD / SynthData::LFO_MAX_FREQ / _lfoWaveSize;
	if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
	int _lfoTimerCoef = int(SynthData::LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		
	// EG
	double _rate0Frames = SynthData::EG_REFERENCE_TIME * SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egPow2tIdx = SynthData::toPow2tIndexValue(SynthData::EG_MAX_CNT / _rate0Frames);
	double _sndOffFrames = SynthData::EG_SOUND_OFF_TIME / SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egSndOff = int(SynthData::EG_MAX_CNT / _sndOffFrames + 0.5);
	// PG
	int _smpl = SynthData::toPow2tIndexValue(_sampleRateD);
	int _pgSamplePow2tIdx = SynthData::toPow2tIndexValue(double(1 << PG_PHASE_FP)) - _smpl;
	// PG WaveTable
	int _pgKeyPow2tIdx = SynthData::toPow2tIndexValue(double(_waveTableSize << PG_PHASE_FP)) - _smpl;
	_pgKeyPow2tIdx += SynthData::toPow2tIndexValue(_referencePitch * std::pow(2.0, -69.0 / 12.0)); // note 0 Pitch
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitWT& _vu = _voiceUnitExt[i];
		_vu.killProcess();
		_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
		_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
		_vu.egPow2tIdx = _egPow2tIdx;
		_vu.egPhaseStepSndOff = _egSndOff;
		_vu.lfoTimerCoef = _lfoTimerCoef;
	}
}
int32 SynthesizerWT::setWaveTableSize(int32 size)
{
	if (size == _waveTableSize) return _waveTableSize;
	int v = 8, _b = 3, _start = _waveTableSize;
	for ( ; v < WAVET_CONT_SIZE; v <<= 1, _b++) if (v >= size) break;
	_waveTableSizeBits = _b;
	_waveTableSize = v;

	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	int32* _wvTbl = WAVE_TABLES;
	for (int i = 0; i < 128; ++i)
	{
		for (int j = _start; j < _waveTableSize; ++j)_wvTbl[j] = _BOTTOM;
		_wvTbl += WAVET_CONT_SIZE;
	}
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitWT& _vu = _voiceUnitExt[i];
		_vu.killProcess();
		_vu.pgPhaseMask = (_waveTableSize << PG_PHASE_FP) - 1;
		_vu.pgPhaseMaskNoise = _vu.pgPhaseMask >> 5;
	}
	_resetKeyParams();
	return _waveTableSize;
}
void SynthesizerWT::getWaveForm(int32 id, int32* buffer, int32 bufferSize, int32 bitDepth)
{
	if (bufferSize > WAVET_CONT_SIZE)bufferSize = WAVET_CONT_SIZE;
	// ---
	if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
	const double _center = double(1 << (bitDepth - 1));
	const int _max = (1 << bitDepth) - 1;
	const int32* _table = &WAVE_TABLES[id << WAVET_CONT_SIZE_BITS];
	const float* _log2vt = SynthData::getInstance()->getLOG2V_TABLE();
	for (int i = 0; i < bufferSize; ++i)
	{
		int v = int(_log2vt[_table[i]] * _center + _center + 0.5);
		if (v > _max) v = _max;
		buffer[i] = v;
	}
}
void SynthesizerWT::setWaveForm(int32 id, const int32* buffer, int32 bufferSize, int32 bitDepth)
{
	if (bufferSize > WAVET_CONT_SIZE)bufferSize = WAVET_CONT_SIZE;
	// ---
	if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
	const int _shift = 8 - bitDepth;
	int32* _table = &WAVE_TABLES[id << WAVET_CONT_SIZE_BITS];
	for (int i = 0; i < bufferSize; ++i)
	{
		int v = buffer[i] << _shift;
		_table[i] = _INDEX_TABLE[v & 0xff];
	}
}
void SynthesizerWT::getWaveFormN(int32 id, float* buffer, int32 bufferSize)
{
	if (bufferSize > WAVET_CONT_SIZE)bufferSize = WAVET_CONT_SIZE;
	// ---
	const int32* _table = &WAVE_TABLES[id << WAVET_CONT_SIZE_BITS];
	const float* _log2vt = SynthData::getInstance()->getLOG2V_TABLE();
	for (int i = 0; i < bufferSize; ++i) buffer[i] = _log2vt[_table[i]];
}
void SynthesizerWT::setWaveFormN(int32 id, const float* buffer, int32 bufferSize, int32 bitDepth)
{
	if (bufferSize > WAVET_CONT_SIZE)bufferSize = WAVET_CONT_SIZE;
	// ---
	if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
	const int _shift = 8 - bitDepth;
	int32* _table = &WAVE_TABLES[id << WAVET_CONT_SIZE_BITS];
	for (int i = 0; i < bufferSize; ++i)
	{
		int v = int(buffer[i] * 127.0f) + 128;
		v >>= _shift;
		_table[i] = _INDEX_TABLE[v];
	}
}
void SynthesizerWT::setSampleData(int32 id, const SampleData& sampleData, bool deepCopy)
{
	allSoundStop();
	SAMPLE_DATAS[id].copy(sampleData, deepCopy);
}
void SynthesizerWT::clearSampleData(int32 id)
{
	allSoundStop();
	SAMPLE_DATAS[id].clear();
}
void SynthesizerWT::factoryDataReset()
{
	if (_voiceUnitExt != nullptr) allSoundStop();
	// waveData
	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	const double _PID2 = 3.141592653589793 * 2.0;
	const double _waveTableSizeD = double(_waveTableSize);
	int _simpleWave[128]{_BOTTOM};
	for (int i = 0; i < _waveTableSize; ++i)
	{
		double _nv = std::sin((i / _waveTableSizeD) * _PID2);
		int _iv = int(_nv * 127.5 - 0.5) + 128;
		_simpleWave[i] = _INDEX_TABLE[_iv];
	}
	if (WAVE_TABLES == nullptr)WAVE_TABLES = new int32[WAVET_CONT_SIZE*128];
	int32* _wvTbl = WAVE_TABLES;
	for (int i = 0; i < 128; ++i)
	{
		for (int j = 0; j < 128; ++j) _wvTbl[j] = _simpleWave[j];
		_wvTbl += WAVET_CONT_SIZE;
	}
	// Preset
	if (TONE_PRESETS == nullptr)TONE_PRESETS = new TonePresetWT[256];
	for (int i = 0; i < 256; ++i)TONE_PRESETS[i].reset();
// Sampler
	if (SAMPLE_DATAS == nullptr)SAMPLE_DATAS = new SampleData[128];
	if (SAMPLER_PRESETS == nullptr)SAMPLER_PRESETS = new SamplerPreset[128];
	for (int i = 0; i < 128; ++i)
	{
		SAMPLE_DATAS[i].clear();
		SAMPLER_PRESETS[i].reset();
	}
}
// Constructor
SynthesizerWT::SynthesizerWT(int32 tblSize, int32 maxPoly, float refPitch, float masVol) : MultiTimberSynthBase::MultiTimberSynthBase(maxPoly, refPitch, masVol)
{
	int v, _b;
	for (v = 8, _b = 3; v < WAVET_CONT_SIZE; v <<= 1, _b++) if (v >= tblSize) break;
	_waveTableSizeBits = _b;
	_waveTableSize = v;
	// create
	if (_INDEX_TABLE == nullptr)_INDEX_TABLE = new int32[256];
	_resetIndexTable();
	factoryDataReset();
	_recreateVU(_maxPolyphony);
	_resetKeyParams();
}
SynthesizerWT::SynthesizerWT(SynthesizerWT && x) noexcept : MultiTimberSynthBase::MultiTimberSynthBase(std::move(x))
{
	// Tables -----
	_INDEX_TABLE = x._INDEX_TABLE; x._INDEX_TABLE = nullptr;
	// Synthesizer - Data -----
	WAVE_TABLES = x.WAVE_TABLES; x.WAVE_TABLES = nullptr;
	SAMPLE_DATAS = x.SAMPLE_DATAS; x.SAMPLE_DATAS = nullptr;
	TONE_PRESETS = x.TONE_PRESETS; x.TONE_PRESETS = nullptr;
	SAMPLER_PRESETS = x.SAMPLER_PRESETS; x.SAMPLER_PRESETS = nullptr;
	_waveTableSizeBits = x._waveTableSizeBits;
	_waveTableSize = x._waveTableSize;
	// Synthesizer - main -----
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;
}
SynthesizerWT & SynthesizerWT::operator=(SynthesizerWT && x) noexcept
{
	if (this == &x)return *this;
	// MultiTimberSynthBase::operator=(std::move(x));
// MultiTimberSynthBase ===
	// Synth Status
	midiInPorts = x.midiInPorts;
	midiOutPorts = x.midiOutPorts;
	receiveSysEx = x.receiveSysEx;
	// Channel Paramters
	if (VOICE_PARTS != nullptr)delete[] VOICE_PARTS;
	VOICE_PARTS = x.VOICE_PARTS; x.VOICE_PARTS = nullptr;
	_sampleRate = x._sampleRate;
	_referencePitch = x._referencePitch;
	_masterVolume = x._masterVolume;
	// VoiceUnits
	_maxPolyphony = x._maxPolyphony;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = x._voiceUnits; x._voiceUnits = nullptr;
	_nton = x._nton;
	_ntonM = x._ntonM;
	// RPN,NRPN -----
	_rpnCtrlFlag = x._rpnCtrlFlag;
	_nrpnCtrlFlag = x._nrpnCtrlFlag;
// SynthesizerWT ===
	// Tables -----
	if (_INDEX_TABLE != nullptr)delete[] _INDEX_TABLE;
	_INDEX_TABLE = x._INDEX_TABLE; x._INDEX_TABLE = nullptr;
	// Synthesizer - Data -----
	if (WAVE_TABLES != nullptr)delete[] WAVE_TABLES;
	WAVE_TABLES = x.WAVE_TABLES; x.WAVE_TABLES = nullptr;
	if (SAMPLE_DATAS != nullptr)delete[] SAMPLE_DATAS;
	SAMPLE_DATAS = x.SAMPLE_DATAS; x.SAMPLE_DATAS = nullptr;
	if (TONE_PRESETS != nullptr)delete[] TONE_PRESETS;
	TONE_PRESETS = x.TONE_PRESETS; x.TONE_PRESETS = nullptr;
	if (SAMPLER_PRESETS != nullptr)delete[] SAMPLER_PRESETS;
	SAMPLER_PRESETS = x.SAMPLER_PRESETS; x.SAMPLER_PRESETS = nullptr;
	_waveTableSizeBits = x._waveTableSizeBits;
	_waveTableSize = x._waveTableSize;
	// Synthesizer - main -----
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;

	return *this;
}
SynthesizerWT::~SynthesizerWT()
{
	if (VOICE_PARTS != nullptr)delete[] VOICE_PARTS;
	VOICE_PARTS = nullptr;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = nullptr;
	if (_INDEX_TABLE != nullptr)delete[] _INDEX_TABLE;
	_INDEX_TABLE = nullptr;
	if (WAVE_TABLES != nullptr)delete[] WAVE_TABLES;
	WAVE_TABLES = nullptr;
	if (SAMPLE_DATAS != nullptr)delete[] SAMPLE_DATAS;
	SAMPLE_DATAS = nullptr;
	if (TONE_PRESETS != nullptr)delete[] TONE_PRESETS;
	TONE_PRESETS = nullptr;
	if (SAMPLER_PRESETS != nullptr)delete[] SAMPLER_PRESETS;
	SAMPLER_PRESETS = nullptr;
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = nullptr;
}
// MidiEvent ---
void SynthesizerWT::_noteOn(const MidiEvent& ev)
{
	VoicePart& _pt = VOICE_PARTS[ev.channel];
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitWT& _vu = _voiceUnitExt[i];
		if (_vu.state == 0)
		{
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	const int _len = _nton + _maxPolyphony;
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitWT& _vu = _voiceUnitExt[i & _ntonM];
		// (KeyOff || SoundOff) -->
		if (_vu.state & 0x60)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitWT& _vu = _voiceUnitExt[i & _ntonM];
		// KeyOn -->
		if (_vu.state & 0x10)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
}
void SynthesizerWT::_readSysExData(const MidiEvent& ev)
{
	if (ev.dataSize < 8)return;
	const uint8* dat = ev.data;
	const uint8* dataEnd = dat + ev.dataSize - 1;
	if (*(dat++) != 0x7Du)return;  // ManuFactureID
	if (*(dat++) != 0x10u)return;  // DeviceID
	if (*(dat++) != 0x20u)return;  // ModelID

	uint8 _padr = 0;
	switch (*(dat++))	// Address 1
	{
	// Common
		case 0x00u:
			if (*(dat++) != 0x00u)return; // Address 2
			switch (*(dat++))	// Address 3
			{
				case 0x00u : setMasterVolume(_MASTER_VOLUME_TABLE[*(dat++)]); if (dat >= dataEnd) break;
				case 0x01u : setReferencePitch(440.0f + (*dat - 64.0f) * 0.5f); break;
			}
			break;
	// Inst (Drum map)
		case 0x05u: _padr = 128u;
	// Inst (Program map)
		case 0x04u: {
			TonePresetWT& _preset = TONE_PRESETS[_padr | *(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _preset.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _preset.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _preset.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _preset.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _preset.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _preset.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _preset.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _preset.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _preset.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _preset.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _preset.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _preset.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _preset.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _preset.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _preset.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _preset.toneName[15] = *(dat++);  if (dat >= dataEnd) break;
				
				case 0x10u : _preset.key     = *(dat++);       if (dat >= dataEnd) break;
				case 0x11u : _preset.pan     = *(dat++);       if (dat >= dataEnd) break;
				case 0x12u : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x13u : _preset.lfoFreq = _preset.lfoFreq & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x14u : _preset.strc    = *(dat++) & 3;                             if (dat >= dataEnd) break;
				case 0x15u : _preset.tune    = _preset.tune & 0x3f80 | *(dat++);         if (dat >= dataEnd) break;
				case 0x16u : _preset.tune    = _preset.tune & 0x007f | *(dat++) << 7;    if (dat >= dataEnd) break;
				case 0x17u : _preset.lms     = *(dat++) & 15;  if (dat >= dataEnd) break;
				case 0x18u : _preset.wave    = *(dat++);       if (dat >= dataEnd) break;
				case 0x19u : _preset.tl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Au : _preset.ar      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Bu : _preset.dr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Cu : _preset.sl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Du : _preset.rr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x1Eu : _preset.kf      = *(dat++) & 3;   if (dat >= dataEnd) break;
				case 0x1Fu : _preset.kfStart = *dat;           break;
			}
			} break;
	// Percussion
		case 0x06u: {
			SamplerPreset& _smplPrst = SAMPLER_PRESETS[*(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _smplPrst.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _smplPrst.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _smplPrst.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _smplPrst.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _smplPrst.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _smplPrst.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _smplPrst.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _smplPrst.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _smplPrst.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _smplPrst.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _smplPrst.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _smplPrst.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _smplPrst.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _smplPrst.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _smplPrst.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _smplPrst.toneName[15] = *(dat++);  if (dat >= dataEnd) break;

				case 0x10u : _smplPrst.smplNum = *(dat++);       if (dat >= dataEnd) break;
				case 0x11u : _smplPrst.pan     = *(dat++);       if (dat >= dataEnd) break;
				case 0x12u : _smplPrst.octv    = *(dat++) & 15;  if (dat >= dataEnd) break;
				case 0x13u : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x14u : _smplPrst.tune    = _smplPrst.tune & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x15u : _smplPrst.tl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x16u : _smplPrst.ar      = *(dat++);       if (dat >= dataEnd) break;
				case 0x17u : _smplPrst.dr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x18u : _smplPrst.sl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x19u : _smplPrst.rr      = *dat;           break;
			}
			} break;
	// Data
		case 0x08u: {
			uint32 _wadr = uint32(*(dat++)) << 7u;
			int32* _wave = WAVE_TABLES + _wadr; // Address 2 [ ID ]
			for (uint8 i = *(dat++); i < 128u; ++i)
			{
				_wave[i] = _INDEX_TABLE[*(dat++) << 1u];
				if (dat >= dataEnd) break;
			}
			} break;
	}
}
/***********************************************
	Class	SynthesizerWT :: VoiceUnitWT
************************************************/
void SynthesizerWT::VoiceUnitWT::keyOff()
{
	// KeyOn --->
	if (state & 0x0010)
	{
		// ATTACK ->
		if (_egState == 1)
		{
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			if (_egCount >= _EG_MAX_CNT)_egCount = 0;
			else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			if (_egCount > _egCntEnvBtm){ killProcess(); return; }
		}
		// DECAY,SUSTAIN ->
		_egGate = _egCntEnvBtm;
		_egCountStep = _egCountStepRR;
		_egState = 8;  // Release
		state = 0x0021;
		// sampler ---
		_loopE = _plbkE;
		_loopF = _loopFk;
	}
}
void SynthesizerWT::VoiceUnitWT::soundOff()
{
	// KeyOn || KeyOff --->
	if (state & 0x0030)
	{
		// ATTACK ->
		if (_egState == 1)
		{
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			if (_egCount >= _EG_MAX_CNT)_egCount = 0;
			else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << (_EG_CNT_FP - 1);
			if (_egCount > _egCntEnvBtm){ killProcess(); return; }
		}
		// DECAY,SUSTAIN ->
		_egGate = _egCntEnvBtm;
		_egCountStep = egPhaseStepSndOff;
		_egState = 0;  // SoundOff, EG OFF
		partNumber = -1; // SoundOff
		state = 0x0041;
		// sampler ---
		_loopE = _plbkE;
		_loopF = _loopFk;
	}
}
int32 SynthesizerWT::VoiceUnitWT::_shiftEGState()
{
	switch (_egState)
	{
		// ATTACK ->
		case 1:
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
			// -> DECAY
			if (_egSL > 0)
			{
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				return _egState = 2;
			}
			// -> SUSTAIN
			_egGate = 0x7fffffff;
			_egCountStep = 0;
			return _egState = 4;
		// DECAY ->
		case 2:
			// -> SUSTAIN
			if ((_egTL + _egSL) < _EG_SUSTAIN_BTM)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				_egGate = 0x7fffffff;
				_egCountStep = 0;
				_egCount = _egCntEnvSus;
				return _egState = 4;
			}
		// DECAY,RELEASE,SoundOFF -> EG OFF
		default: return _egState = 0;
	}
}
void SynthesizerWT::VoiceUnitWT::keyOn(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0x0011;
	partNumber = ch;
	noteNumber = note;
// Part, Preset ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	TonePresetWT* _preset = nullptr;
	int _keyFlw = egPow2tIdx;
	if (_part.mode == 2)
	{
		_preset = &TONE_PRESETS[noteNumber | 128];
		_pgKey = _POW2T_KEY_TABLE[_preset->key] + (_preset->tune - 8192) + pgKeyPow2tIdx;
		//_keyFlw = 0;
		_prPan = _preset->pan;
	}
	else
	{
		_preset = &TONE_PRESETS[_part.programChange];
		_pgKey = _POW2T_KEY_TABLE[noteNumber] + (_preset->tune - 8192) + pgKeyPow2tIdx;
		_keyFlw += _EG_KEY_TABLE[127 + noteNumber - _preset->kfStart] << _preset->kf;
		_prPan = 64;
	}
// wave ---
	_waveTable = &WAVE_TABLES[_preset->wave << WAVET_CONT_SIZE_BITS];
// LFO ---
	_lfoModSens = _preset->lms;
	_lfoMod = _part.ccValues[1] * _lfoModSens;
	_lfoTimerStep = _preset->lfoFreq * lfoTimerCoef;
	_lfoTimer = _LFO_TIMER_INIT_VALUE;
	_lfoPhase = _preset->strc + 1 & 1;
	_lfoOutPM = 0;
	_lfoOutAM = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// PG ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v = _pgKey + _pgBend;// & 0x7fffffff;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase = 0;
// EG ---
	_egTL = _preset->tl;
	_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
	_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
	_egSL = _preset->sl;
	_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
	if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
	v = (_preset->dr << _EG_PRAM_OCTV_LS) + _keyFlw;
	_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset->rr << _EG_PRAM_OCTV_LS) + _keyFlw;
	_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset->ar < 127) {
		_egCount = _EG_CNT_ENV_ATK;
		_egLevelTable = _EG_LEVEL_TABLE_UP;
		_egGate = _EG_MAX_CNT;
		v = (_preset->ar << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState = 1;
	} else {
		_egLevelTable = _EG_LEVEL_TABLE_DW;
		_egCount = 0;
	// -> DECAY
		if (_egSL > 0) {
			_egGate = _egCntEnvSus;
			_egCountStep = _egCountStepDR;
			_egState = 2;
	// -> SUSTAIN
		} else {
			_egGate = 0x7fffffff;  // dummy
			_egCountStep = 0;  // dummy
			_egState = 4; }
	}
// process ---
	process = _PROC_PTRS[_preset->strc];
}
void SynthesizerWT::VoiceUnitWT::keyOnSampler(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0x0011;
	partNumber = ch;
	noteNumber = note;
// Part, Preset ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	SamplerPreset& _preset = SAMPLER_PRESETS[noteNumber];
	SampleData& _sample = SAMPLE_DATAS[_preset.smplNum];
	if (_sample.sampleData == nullptr){ killProcess(); return; }
	_waveTable = _sample.sampleData;
	_plbkE = _sample.plbkE;
	_loopE = _sample.loopE;
	_loopF = _sample.loopF;
	_loopFk = _sample.loopFk;
// LFO ---
	_lfoOutPM = 0;
	_lfoOutAM = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_prPan = _preset.pan;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// PG ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	_pgKey = ((_preset.octv - 8) << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
	int v = _pgKey + _pgBend;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase = _sample.plbkS;
	_pgPhaseAFP = 0;
// EG ---
	_egTL = _preset.tl;
	_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
	_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
	_egSL = _preset.sl;
	_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
	if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
	v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset.ar < 127) {
		_egCount = _EG_CNT_ENV_ATK;
		_egLevelTable = _EG_LEVEL_TABLE_UP;
		_egGate = _EG_MAX_CNT;
		v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState = 1;
	} else {
		_egLevelTable = _EG_LEVEL_TABLE_DW;
		_egCount = 0;
	// -> DECAY
		if (_egSL > 0) {
			_egGate = _egCntEnvSus;
			_egCountStep = _egCountStepDR;
			_egState = 2;
	// -> SUSTAIN
		} else {
			_egGate = 0x7fffffff;  // dummy
			_egCountStep = 0;  // dummy
			_egState = 4; }
	}
// process ---
	process = (_sample.channelCount == 2)? _PROC_PTRS[5] : _PROC_PTRS[4];
}
void SynthesizerWT::VoiceUnitWT::updPartBendSens(const int32& val)
{
	_ptBendSens = _POW2T_KEY_TABLE[val];
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v = _pgKey + _pgBend + _lfoOutPM;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
}
void SynthesizerWT::VoiceUnitWT::updPartPitchBend(const int32& val)
{
	_ptBend = val - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v = _pgKey + _pgBend + _lfoOutPM;
	_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
}
void SynthesizerWT::VoiceUnitWT::updPartModulation(const int32& val)
{
	_lfoMod = val * _lfoModSens;
}
void SynthesizerWT::VoiceUnitWT::updateLfoFrequency(const int32& val)
{
	_lfoTimerStep = val * lfoTimerCoef;
}
void SynthesizerWT::VoiceUnitWT::_proc0(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
	// pg update --->
		v = _wt[_pgPhase >> PG_PHASE_FP] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhase = _pgPhase + _pgPhaseStep & pgPhaseMask;
	}
}
void SynthesizerWT::VoiceUnitWT::_proc1(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
			int l = _pgKey + _pgBend + _lfoOutPM;
			_pgPhaseStep = _POW2_TABLE[l & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[l >> _POW2T_OCTV_BITS];
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
	// pg update --->
		v = _wt[_pgPhase >> PG_PHASE_FP] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
		_egCount += _egCountStep;
		_pgPhase = _pgPhase + _pgPhaseStep & pgPhaseMask;
	}
}
void SynthesizerWT::VoiceUnitWT::_proc2(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			int l = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;
			_lfoOutAM = l & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
	// pg update --->
		v = _wt[_pgPhase >> PG_PHASE_FP] + v + _lfoOutAM;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
		_egCount += _egCountStep;
		_pgPhase = _pgPhase + _pgPhaseStep & pgPhaseMask;
	}
}
void SynthesizerWT::VoiceUnitWT::_proc3(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
	// pg update --->
		//while (_pgPhase > _PG_PHASE_FP_MASK)
		if (_pgPhase > pgPhaseMaskNoise)
		{
			//_pgPhase -= 1 << _PG_PHASE_FP;
			_pgPhase &= pgPhaseMaskNoise;
			_lfsrReg = _lfsrReg >> 1 ^ -(_lfsrReg & 1) & _LFSR_TAP_15;
			_noise = _lfsrReg & 1;
		}
		float n = _log2vT[_noise + v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhase += _pgPhaseStep;
	}
}
void SynthesizerWT::VoiceUnitWT::_proc4(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerWT::VoiceUnitWT::_proc5(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount >= _egGate)
		{
			if (_shiftEGState())_et = _egLevelTable;
			else { killProcess(); return; }
		}
		int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
	// pg update --->
		if (_pgPhase >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_egCount += _egCountStep;
		_pgPhaseAFP += _pgPhaseStep;
		_pgPhase += _pgPhaseAFP >> PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
// Constructor
SynthesizerWT::VoiceUnitWT::VoiceUnitWT() : VoiceUnitBase()
{
	SynthData* _synthData = SynthData::getInstance();
	// Log2 Volume Table
	_LOG2V_TABLE = _synthData->getLOG2V_TABLE();
	// Pow2 Table
	_POW2_TABLE      = _synthData->getPOW2_TABLE();
	_POW2T_RS_TABLE  = _synthData->getPOW2T_RS_TABLE();
	_POW2T_KEY_TABLE = _synthData->getPOW2T_KEY_TABLE();
	// EG
	_EG_TOTAL_LEVEL_TABLE = _synthData->getEG_TOTAL_LEVEL_TABLE();
	_EG_LEVEL_TABLE_UP    = _synthData->getEG_LEVEL_TABLE_UP();
	_EG_LEVEL_TABLE_DW    = _synthData->getEG_LEVEL_TABLE_DW();
	_EG_KEY_TABLE         = _synthData->getEG_KEY_TABLE();
	_EG_POW2T_RS_TABLE    = _synthData->getEG_POW2T_RS_TABLE();
	_EG_CNT_GATE_TABLE    = _synthData->getEG_CNT_GATE_TABLE();
	// etc
	_LFO_WAVE_TABLE = &_synthData->getLFO_WAVE_TABLES()[0];
}

}}}
